from typing import Dict, Any

from sqlmodel import Field, SQLModel
from sqlalchemy import Column, JSON
from bullish.analysis.analysis import Analysis
from bullish.analysis.filter import FilteredResults
from bullish.jobs.models import JobTracker
from sqlalchemy import Index


class BaseTable(SQLModel):
    symbol: str = Field(primary_key=True)
    source: str = Field(primary_key=True)


dynamic_indexes = tuple(
    Index(f"ix_analysis_{col}", col) for col in Analysis.model_fields
)


class AnalysisORM(BaseTable, Analysis, table=True):
    __tablename__ = "analysis"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012


AnalysisORM.__table_args__ = tuple(  # type: ignore # noqa: RUF005
    Index(f"ix_{AnalysisORM.__tablename__}_{col.name}", col)
    for col in AnalysisORM.__table__.columns
    if not col.primary_key and not col.index and col.name != "id"
) + (AnalysisORM.__table_args__,)


class JobTrackerORM(SQLModel, JobTracker, table=True):
    __tablename__ = "jobtracker"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012
    job_id: str = Field(primary_key=True)
    type: str  # type: ignore
    status: str  # type: ignore


class FilteredResultsORM(SQLModel, FilteredResults, table=True):
    __tablename__ = "filteredresults"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012
    name: str = Field(primary_key=True)
    symbols: list[str] = Field(sa_column=Column(JSON))
    filter_query: Dict[str, Any] = Field(sa_column=Column(JSON))  # type: ignore



import datetime
from typing import Dict, Any, Optional

from bullish.analysis.filter import FilterQuery
from pydantic import BaseModel, Field

DATE_THRESHOLD = [
    datetime.date.today() - datetime.timedelta(days=10),
    datetime.date.today(),
]


class NamedFilterQuery(FilterQuery):
    name: str
    description: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump(
            exclude_unset=True,
            exclude_none=True,
            exclude_defaults=True,
            exclude={"name"},
        )


STRONG_FUNDAMENTALS = NamedFilterQuery(
    name="Strong Fundamentals",
    income=[
        "positive_operating_income",
        "growing_operating_income",
        "positive_net_income",
        "growing_net_income",
    ],
    cash_flow=["positive_free_cash_flow", "growing_operating_cash_flow"],
    eps=["positive_diluted_eps", "growing_diluted_eps"],
    properties=[
        "operating_cash_flow_is_higher_than_net_income",
        "positive_return_on_equity",
        "positive_return_on_assets",
        "positive_debt_to_equity",
    ],
    market_capitalization=[1e10, 1e12],  # 1 billion to 1 trillion
    rsi_bullish_crossover_30=DATE_THRESHOLD,
)

GOOD_FUNDAMENTALS = NamedFilterQuery(
    name="Good Fundamentals",
    income=[
        "positive_operating_income",
        "positive_net_income",
    ],
    cash_flow=["positive_free_cash_flow"],
    eps=["positive_diluted_eps"],
    properties=[
        "positive_return_on_equity",
        "positive_return_on_assets",
        "positive_debt_to_equity",
    ],
    market_capitalization=[1e10, 1e12],  # 1 billion to 1 trillion
    rsi_bullish_crossover_30=DATE_THRESHOLD,
)


SHOOTING_STARS = NamedFilterQuery(
    name="Shooting stars",
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    market_capitalization=[1e9, 1e12],  # 1 billion to 1 trillion
    order_by_desc="median_yearly_growth",
    order_by_asc="last_price",
)

RSI_CROSSOVER_TECH = NamedFilterQuery(
    name="RSI cross-over",
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    return_after_rsi_crossover_45_period_90=[0.0, 100],
    rsi_bullish_crossover_45=DATE_THRESHOLD,
    market_capitalization=[5e8, 1e11],  # 1 billion to 1 trillion
    order_by_desc="market_capitalization",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
    industry=[
        "Semiconductors",
        "Software - Application",
        "Software - Infrastructure",
        "Biotechnology",
        "Diagnostics & Research",
        "Medical Devices",
        "Health Information Services",
        "Internet Retail",
        "Electronic Gaming & Multimedia",
        "Internet Content & Information",
        "Solar",
        "Information Technology Services",
        "Scientific & Technical Instruments",
        "Semiconductor Equipment & Materials",
        "Diagnostics & Research",
    ],
)
RSI_CROSSOVER_TECH_PE = NamedFilterQuery(
    name="RSI cross-over P/E",
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[5, 30],  # P/E ratio between 10 and 100
    rsi_bullish_crossover_45=DATE_THRESHOLD,
    market_capitalization=[5e8, 1e12],  # 1 billion to 1 trillion
    order_by_desc="market_capitalization",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
    industry=[
        "Semiconductors",
        "Software - Application",
        "Software - Infrastructure",
        "Biotechnology",
        "Diagnostics & Research",
        "Medical Devices",
        "Health Information Services",
        "Internet Retail",
        "Electronic Gaming & Multimedia",
        "Internet Content & Information",
        "Solar",
        "Information Technology Services",
        "Scientific & Technical Instruments",
        "Semiconductor Equipment & Materials",
        "Diagnostics & Research",
    ],
)

MICRO_CAP_EVENT_SPECULATION = NamedFilterQuery(
    name="Micro-Cap Event Speculation",
    description="seeks tiny names where unusual volume and price gaps hint at "
    "pending corporate events (patent win, FDA news, buy-out rumors).",
    positive_adosc_20_day_breakout=DATE_THRESHOLD,
    rate_of_change_30=[20, 100],  # 10% to 50% in the last 30 days
    market_capitalization=[0, 5e8],
)

MOMENTUM_BREAKOUT_HUNTER = NamedFilterQuery(
    name="Momentum Breakout Hunter",
    description="A confluence of medium-term (50/200 MA) and "
    "shorter oscillators suggests fresh upside momentum with fuel left.",
    income=[
        "positive_operating_income",
        "positive_net_income",
    ],
    cash_flow=["positive_free_cash_flow"],
    golden_cross=DATE_THRESHOLD,
    adx_14_long=DATE_THRESHOLD,
    rate_of_change_30=[0, 100],
    rsi_neutral=DATE_THRESHOLD,
)

DEEP_VALUE_PLUS_CATALYST = NamedFilterQuery(
    name="Deep-Value Plus Catalyst",
    description="Seeks beaten-down names that just printed a bullish "
    "candle and early accumulation signals—often the first leg of a bottom.",
    income=[
        "positive_operating_income",
        "positive_net_income",
    ],
    lower_than_200_day_high=DATE_THRESHOLD,
    rate_of_change_30=[3, 100],
    rsi_bullish_crossover_30=DATE_THRESHOLD,
)
END_OF_TREND_REVERSAL = NamedFilterQuery(
    name="End of trend reversal",
    description="Layers long-term MA breach with momentum exhaustion and a "
    "bullish candle—classic setup for mean-reversion traders.",
    death_cross=DATE_THRESHOLD,
    rsi_oversold=DATE_THRESHOLD,
    candlesticks=["cdlmorningstart", "cdlabandonedbaby", "cdl3whitesoldiers"],
)

HIGH_QUALITY_CASH_GENERATOR = NamedFilterQuery(
    name="High Quality Cash Generator",
    description="This quartet isolates companies that are profitable, cash-rich, and disciplined with leverage. "
    "Ideal first pass for “quality” or “compounder” "
    "portfolios where downside protection matters as much as upside.",
    income=[
        "positive_net_income",
    ],
    cash_flow=["positive_free_cash_flow"],
    properties=[
        "operating_cash_flow_is_higher_than_net_income",
        "positive_return_on_equity",
        "positive_return_on_assets",
        "positive_debt_to_equity",
    ],
)

EARNINGS_ACCELERATION_TREND_CONFIRMATION = NamedFilterQuery(
    name="Earnings Acceleration Trend Confirmation",
    description="Pairs fundamental acceleration with momentum confirmation. Research shows this “double positive” "
    "outperforms simple momentum because it filters out purely sentiment-driven rallies.",
    income=[
        "growing_operating_income",
        "positive_net_income",
    ],
    eps=["growing_basic_eps"],
    golden_cross=DATE_THRESHOLD,
    macd_12_26_9_bullish_crossover=DATE_THRESHOLD,
    adx_14_long=DATE_THRESHOLD,
)
DIVIDEND_GROWTH_COMPOUNDER = NamedFilterQuery(
    name="Dividend-Growth Compounders",
    description="Separates true dividend growers from high-yield traps. "
    "Critical for income portfolios that need both yield and growth to beat inflation.",
    mean_dividend_payout_ratio=[0, 0.6],  # 0% to 60% payout ratio
    cash_flow=[
        "positive_free_cash_flow",
        "quarterly_positive_free_cash_flow",
        "growing_operating_cash_flow",
    ],
    properties=["quarterly_positive_return_on_equity"],
)

BREAK_OUT_MOMENTUM = NamedFilterQuery(
    name="Break-out Momentum",
    description="Combines price, volume, and pattern confirmation. Great for tactical traders seeking "
    "quick continuation moves with statistically higher follow-through.",
    adosc_crosses_above_0=DATE_THRESHOLD,
    positive_adosc_20_day_breakout=DATE_THRESHOLD,
    rsi_bullish_crossover_30=DATE_THRESHOLD,
)

OVERSOLD_MEAN_REVERSION = NamedFilterQuery(
    name="Oversold Mean Reversion",
    description="Gives contrarian traders a high-probability bounce setup by "
    "stacking three different oversold measures plus a reversal pattern.",
    rsi_oversold=DATE_THRESHOLD,
    stoch_oversold=DATE_THRESHOLD,
    mfi_oversold=DATE_THRESHOLD,
    lower_than_200_day_high=DATE_THRESHOLD,
)
RSI_CROSSOVER_30_GROWTH_STOCK_STRONG_FUNDAMENTAL = NamedFilterQuery(
    name="RSI cross-over 30 growth stock strong fundamental",
    income=[
        "positive_operating_income",
        "growing_operating_income",
        "positive_net_income",
        "growing_net_income",
    ],
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[20, 40],
    rsi_bullish_crossover_30=DATE_THRESHOLD,
    market_capitalization=[5e8, 1e12],
    order_by_desc="market_capitalization",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
RSI_CROSSOVER_40_GROWTH_STOCK_STRONG_FUNDAMENTAL = NamedFilterQuery(
    name="RSI cross-over 40 growth stock strong fundamental",
    income=[
        "positive_operating_income",
        "growing_operating_income",
        "positive_net_income",
        "growing_net_income",
    ],
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[20, 40],
    rsi_bullish_crossover_40=DATE_THRESHOLD,
    market_capitalization=[5e8, 1e12],
    order_by_desc="market_capitalization",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
RSI_CROSSOVER_45_GROWTH_STOCK_STRONG_FUNDAMENTAL = NamedFilterQuery(
    name="RSI cross-over 45 growth stock strong fundamental",
    income=[
        "positive_operating_income",
        "growing_operating_income",
        "positive_net_income",
        "growing_net_income",
    ],
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[20, 40],
    rsi_bullish_crossover_45=DATE_THRESHOLD,
    market_capitalization=[5e8, 1e12],
    order_by_desc="market_capitalization",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
RSI_CROSSOVER_30_GROWTH_STOCK = NamedFilterQuery(
    name="RSI cross-over 30 growth stock",
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[20, 40],
    rsi_bullish_crossover_30=DATE_THRESHOLD,
    market_capitalization=[5e8, 1e12],
    order_by_desc="market_capitalization",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
RSI_CROSSOVER_40_GROWTH_STOCK = NamedFilterQuery(
    name="RSI cross-over 40 growth stock",
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[20, 40],
    rsi_bullish_crossover_40=DATE_THRESHOLD,
    market_capitalization=[5e8, 1e12],
    order_by_desc="market_capitalization",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
RSI_CROSSOVER_45_GROWTH_STOCK = NamedFilterQuery(
    name="RSI cross-over 45 growth stock",
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[20, 40],
    rsi_bullish_crossover_45=DATE_THRESHOLD,
    market_capitalization=[5e8, 1e12],
    order_by_desc="market_capitalization",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
MOMENTUM_STOCK_STRONG_FUNDAMENTAL = NamedFilterQuery(
    name="Momentum stock strong fundamental",
    income=[
        "positive_operating_income",
        "growing_operating_income",
        "positive_net_income",
        "growing_net_income",
    ],
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[10, 400],
    last_price=[1, 70],
    order_by_desc="momentum",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
MOMENTUM_STOCK = NamedFilterQuery(
    name="Momentum stock",
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[10, 400],
    last_price=[1, 70],
    order_by_desc="momentum",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
MOMENTUM_STOCK_NO_FUNDAMENTAL_CHECKS = NamedFilterQuery(
    name="Momentum stock no fundamental checks",
    price_per_earning_ratio=[10, 500],
    last_price=[1, 10000],
    order_by_desc="momentum",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
MOMENTUM_TIME_SPAN_1_MONTH = NamedFilterQuery(
    name="Momentum 1 month",
    price_per_earning_ratio=[10, 500],
    last_price=[1, 10000],
    momentum_time_span=[
        datetime.date.today() - datetime.timedelta(days=90),
        datetime.date.today() - datetime.timedelta(days=31),
    ],
    macd_12_26_9_bullish_crossover=[
        datetime.date.today() - datetime.timedelta(days=10),
        datetime.date.today(),
    ],
    order_by_desc="momentum",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)
MOMENTUM_TIME_SPAN_1_MONTH_STRONG_FUNDAMENTALS = NamedFilterQuery(
    name="Momentum 1 month strong fundamentals",
    income=[
        "positive_operating_income",
        "growing_operating_income",
        "positive_net_income",
        "growing_net_income",
    ],
    cash_flow=["positive_free_cash_flow"],
    properties=["operating_cash_flow_is_higher_than_net_income"],
    price_per_earning_ratio=[10, 500],
    last_price=[1, 10000],
    momentum_time_span=[
        datetime.date.today() - datetime.timedelta(days=90),
        datetime.date.today() - datetime.timedelta(days=31),
    ],
    macd_12_26_9_bullish_crossover=[
        datetime.date.today() - datetime.timedelta(days=10),
        datetime.date.today(),
    ],
    order_by_desc="momentum",
    country=["Germany", "United states", "France", "United kingdom", "Canada", "Japan"],
)


def predefined_filters() -> list[NamedFilterQuery]:
    return [
        RSI_CROSSOVER_30_GROWTH_STOCK_STRONG_FUNDAMENTAL,
        RSI_CROSSOVER_40_GROWTH_STOCK_STRONG_FUNDAMENTAL,
        RSI_CROSSOVER_45_GROWTH_STOCK_STRONG_FUNDAMENTAL,
        RSI_CROSSOVER_30_GROWTH_STOCK,
        RSI_CROSSOVER_40_GROWTH_STOCK,
        RSI_CROSSOVER_45_GROWTH_STOCK,
        MOMENTUM_STOCK_STRONG_FUNDAMENTAL,
        MOMENTUM_STOCK,
        MOMENTUM_STOCK_NO_FUNDAMENTAL_CHECKS,
        MOMENTUM_TIME_SPAN_1_MONTH,
        MOMENTUM_TIME_SPAN_1_MONTH_STRONG_FUNDAMENTALS,
    ]


class PredefinedFilters(BaseModel):
    filters: list[NamedFilterQuery] = Field(default_factory=predefined_filters)

    def get_predefined_filter_names(self) -> list[str]:
        return [filter.name for filter in self.filters]

    def get_predefined_filter(self, name: str) -> Dict[str, Any]:
        for filter in self.filters:
            if filter.name == name:
                return filter.to_dict()
        raise ValueError(f"Filter with name '{name}' not found.")



import json
import logging
from datetime import date
from functools import cached_property
from pathlib import Path
from sqlite3 import OperationalError
from typing import TYPE_CHECKING, Any, List, Optional

import pandas as pd
from bearish.database.crud import BearishDb  # type: ignore
from bearish.models.base import Ticker  # type: ignore
from bearish.database.schemas import EarningsDateORM  # type: ignore
from pydantic import ConfigDict
from sqlalchemy import Engine, create_engine, insert, delete, update
from sqlmodel import Session, select

from bullish.analysis.analysis import Analysis
from bullish.database.schemas import AnalysisORM, JobTrackerORM, FilteredResultsORM
from bullish.database.scripts.upgrade import upgrade
from bullish.exceptions import DatabaseFileNotFoundError
from bullish.analysis.filter import FilteredResults
from bullish.interface.interface import BullishDbBase
from bullish.jobs.models import JobTracker, JobTrackerStatus

if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)

BATCH_SIZE = 5000


class BullishDb(BearishDb, BullishDbBase):  # type: ignore
    model_config = ConfigDict(arbitrary_types_allowed=True)
    database_path: Path

    def valid(self) -> bool:
        """Check if the database is valid."""
        return self.database_path.exists() and self.database_path.is_file()

    @cached_property
    def _engine(self) -> Engine:
        if not self.valid():
            raise DatabaseFileNotFoundError("Database file not found.")
        database_url = f"sqlite:///{Path(self.database_path)}"
        try:
            upgrade(self.database_path)
        except OperationalError as e:
            logger.warning(
                f"Failed to upgrade the database at {self.database_path}. "
                f"Reason: {e}"
                "Skipping upgrade. "
            )
        engine = create_engine(database_url)
        return engine

    def model_post_init(self, __context: Any) -> None:
        self._engine  # noqa: B018

    def _write_analysis(self, analysis: Analysis) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(AnalysisORM)
                .prefix_with("OR REPLACE")
                .values(analysis.model_dump())
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def _write_many_analysis(self, many_analysis: List[Analysis]) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(AnalysisORM)
                .prefix_with("OR REPLACE")
                .values([a.model_dump() for a in many_analysis])
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def _read_analysis(self, ticker: Ticker) -> Optional[Analysis]:
        with Session(self._engine) as session:
            query = select(AnalysisORM).where(AnalysisORM.symbol == ticker.symbol)
            analysis = session.exec(query).first()
            if not analysis:
                return None
            return Analysis.model_validate(analysis)

    def read_symbols(self) -> List[str]:
        query = "SELECT DISTINCT symbol FROM analysis"
        data = pd.read_sql_query(query, self._engine)
        return data["symbol"].tolist()

    def _read_analysis_data(
        self, columns: List[str], symbols: Optional[List[str]] = None
    ) -> pd.DataFrame:
        columns_ = ",".join(columns)
        if symbols:
            symbols_str = ",".join([f"'{s}'" for s in symbols])
            query = f"""SELECT {columns_} FROM analysis WHERE symbol IN ({symbols_str})"""  # noqa: S608
        else:
            query = f"""SELECT {columns_} FROM analysis"""  # noqa: S608
        return pd.read_sql_query(query, self._engine)

    def _read_filter_query(self, query: str) -> pd.DataFrame:
        return pd.read_sql(
            query,
            con=self._engine,
        )

    def _read_job_trackers(self) -> pd.DataFrame:
        query = "SELECT * FROM jobtracker ORDER BY started_at DESC"
        return pd.read_sql_query(query, self._engine)

    def write_job_tracker(self, job_tracker: JobTracker) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(JobTrackerORM)
                .prefix_with("OR REPLACE")
                .values(job_tracker.model_dump())
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def read_job_tracker(self, task_id: str) -> Optional[JobTracker]:
        stmt = select(JobTrackerORM).where(JobTrackerORM.job_id == task_id)
        with Session(self._engine) as session:
            result = session.execute(stmt).scalar_one_or_none()
            if result:
                return JobTracker.model_validate(result.model_dump())
            return None

    def delete_job_trackers(self, job_ids: List[str]) -> None:
        with Session(self._engine) as session:
            stmt = delete(JobTrackerORM).where(JobTrackerORM.job_id.in_(job_ids))  # type: ignore
            result = session.execute(stmt)

            if result.rowcount > 0:  # type: ignore
                session.commit()
            else:
                logger.warning(f"Job tracker(s) with ID(s) {job_ids} not found.")

    def update_job_tracker_status(self, job_tracker_status: JobTrackerStatus) -> None:
        with Session(self._engine) as session:
            stmt = (
                update(JobTrackerORM)
                .where(JobTrackerORM.job_id == job_tracker_status.job_id)  # type: ignore
                .values(status=job_tracker_status.status)
            )
            result = session.execute(stmt)

            if result.rowcount > 0:  # type: ignore
                session.commit()
            else:
                logger.warning(
                    f"Job tracker with ID {job_tracker_status.job_id} not found."
                )

    def read_filtered_results(self, name: str) -> Optional[FilteredResults]:
        with Session(self._engine) as session:
            stmt = select(FilteredResultsORM).where(FilteredResultsORM.name == name)
            result = session.execute(stmt).scalar_one_or_none()

            if result:
                return FilteredResults.model_validate(
                    result.model_dump()
                )  # if you're using Pydantic or DTOs
            return None

    def read_list_filtered_results(self) -> List[str]:
        with Session(self._engine) as session:
            stmt = select(FilteredResultsORM.name)
            result = session.execute(stmt).scalars().all()
            return list(result)

    def write_filtered_results(self, filtered_results: FilteredResults) -> None:
        with Session(self._engine) as session:
            data = filtered_results.model_dump_json(
                exclude_unset=True, exclude_defaults=True
            )
            stmt = (
                insert(FilteredResultsORM)
                .prefix_with("OR REPLACE")
                .values(json.loads(data))
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def read_dates(self, symbol: str) -> List[date]:
        with Session(self._engine) as session:
            return [
                r.date()
                for r in session.exec(
                    select(EarningsDateORM.date).where(EarningsDateORM.symbol == symbol)
                )
            ]



import abc
import logging
from datetime import date
from typing import List, Optional

import pandas as pd
from bearish.interface.interface import BearishDbBase  # type: ignore
from bearish.models.base import Ticker  # type: ignore


from bullish.analysis.analysis import Analysis, AnalysisView
from bullish.analysis.filter import FilterQuery, FilteredResults
from bullish.jobs.models import JobTracker, JobTrackerStatus, add_icons

logger = logging.getLogger(__name__)


class BullishDbBase(BearishDbBase):  # type: ignore
    def write_analysis(self, analysis: "Analysis") -> None:
        return self._write_analysis(analysis)

    def write_many_analysis(self, many_analysis: List["Analysis"]) -> None:
        return self._write_many_analysis(many_analysis)

    def read_analysis(self, ticker: Ticker) -> Optional["Analysis"]:
        return self._read_analysis(ticker)

    def read_filter_query(self, query: FilterQuery) -> pd.DataFrame:

        query_ = query.to_query()
        fields = ",".join(list(AnalysisView.model_fields))
        query_str: str = f""" 
        SELECT {fields} FROM analysis WHERE {query_} LIMIT 1000
        """  # noqa: S608
        return self._read_filter_query(query_str)

    def read_analysis_data(
        self, columns: Optional[List[str]] = None, symbols: Optional[List[str]] = None
    ) -> pd.DataFrame:
        columns = columns or list(AnalysisView.model_fields)
        data = self._read_analysis_data(columns, symbols=symbols)
        if set(data.columns) != set(columns):
            raise ValueError(
                f"Expected columns {columns}, but got {data.columns.tolist()}"
            )
        return data

    def read_job_trackers(self) -> pd.DataFrame:
        return add_icons(self._read_job_trackers())

    @abc.abstractmethod
    def _read_job_trackers(self) -> pd.DataFrame: ...

    @abc.abstractmethod
    def write_job_tracker(self, job_tracker: JobTracker) -> None: ...

    @abc.abstractmethod
    def delete_job_trackers(self, job_ids: List[str]) -> None: ...

    @abc.abstractmethod
    def update_job_tracker_status(
        self, job_tracker_status: JobTrackerStatus
    ) -> None: ...

    @abc.abstractmethod
    def _write_analysis(self, analysis: "Analysis") -> None: ...

    @abc.abstractmethod
    def _write_many_analysis(self, many_analysis: List["Analysis"]) -> None: ...

    @abc.abstractmethod
    def _read_analysis(self, ticker: Ticker) -> Optional["Analysis"]: ...

    @abc.abstractmethod
    def _read_filter_query(self, query: str) -> pd.DataFrame: ...

    @abc.abstractmethod
    def _read_analysis_data(
        self, columns: List[str], symbols: Optional[List[str]] = None
    ) -> pd.DataFrame: ...

    @abc.abstractmethod
    def read_filtered_results(self, name: str) -> Optional[FilteredResults]: ...

    @abc.abstractmethod
    def read_list_filtered_results(self) -> List[str]: ...

    @abc.abstractmethod
    def write_filtered_results(self, filtered_results: FilteredResults) -> None: ...

    @abc.abstractmethod
    def read_symbols(self) -> List[str]: ...

    @abc.abstractmethod
    def read_job_tracker(self, task_id: str) -> Optional[JobTracker]: ...

    @abc.abstractmethod
    def read_dates(self, symbol: str) -> List[date]: ...



import logging
from datetime import date
from typing import Optional, List, Callable, Any, Literal, Dict, Union

import numpy as np
import pandas as pd
from pydantic import BaseModel, Field, PrivateAttr, create_model

from bullish.analysis.functions import (
    cross,
    cross_value,
    ADX,
    MACD,
    RSI,
    STOCH,
    MFI,
    ROC,
    CANDLESTOCK_PATTERNS,
    SMA,
    ADOSC,
    PRICE,
    compute_percentile_return_after_rsi_crossover,
    momentum,
)

logger = logging.getLogger(__name__)
SignalType = Literal["Short", "Long", "Oversold", "Overbought", "Value"]


class Signal(BaseModel):
    name: str
    type_info: SignalType
    type: Any
    range: Optional[List[float]] = None
    function: Callable[[pd.DataFrame], Optional[Union[date, float]]]
    description: str
    date: Optional[date] = None
    value: Optional[float] = None

    def is_date(self) -> bool:
        if self.type == Optional[date]:
            return True
        elif self.type == Optional[float]:
            return False
        else:
            raise NotImplementedError

    def compute(self, data: pd.DataFrame) -> None:
        if self.is_date():
            self.date = self.function(data)  # type: ignore
        else:
            self.value = self.function(data)  # type: ignore


class Indicator(BaseModel):
    name: str
    description: str
    expected_columns: List[str]
    function: Callable[[pd.DataFrame], pd.DataFrame]
    _data: pd.DataFrame = PrivateAttr(default=pd.DataFrame())
    signals: List[Signal] = Field(default_factory=list)

    def compute(self, data: pd.DataFrame) -> None:
        results = self.function(data)
        if not set(self.expected_columns).issubset(results.columns):
            raise ValueError(
                f"Expected columns {self.expected_columns}, but got {results.columns.tolist()}"
            )
        self._data = results
        self._signals()

    def _signals(self) -> None:
        for signal in self.signals:
            try:
                signal.compute(self._data)
            except Exception as e:  # noqa: PERF203
                logger.error(
                    f"Fail to compute signal {signal.name} for indicator {self.name}: {e}"
                )


def indicators_factory() -> List[Indicator]:
    return [
        Indicator(
            name="ADX_14",
            description="Average Directional Movement Index",
            expected_columns=["ADX_14", "MINUS_DI", "PLUS_DI"],
            function=ADX.call,
            signals=[
                Signal(
                    name="ADX_14_LONG",
                    description="ADX 14 Long Signal",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d[
                        (d.ADX_14 > 20) & (d.PLUS_DI > d.MINUS_DI)
                    ].last_valid_index(),
                ),
                Signal(
                    name="ADX_14_SHORT",
                    description="ADX 14 Short Signal",
                    type_info="Short",
                    type=Optional[date],
                    function=lambda d: d[
                        (d.ADX_14 > 20) & (d.MINUS_DI > d.PLUS_DI)
                    ].last_valid_index(),
                ),
            ],
        ),
        Indicator(
            name="MACD_12_26_9",
            description="Moving Average Convergence/Divergence",
            expected_columns=[
                "MACD_12_26_9",
                "MACD_12_26_9_SIGNAL",
                "MACD_12_26_9_HIST",
            ],
            function=MACD.call,
            signals=[
                Signal(
                    name="MACD_12_26_9_BULLISH_CROSSOVER",
                    description="MACD 12-26-9 Bullish Crossover",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross(d.MACD_12_26_9, d.MACD_12_26_9_SIGNAL),
                ),
                Signal(
                    name="MACD_12_26_9_BEARISH_CROSSOVER",
                    description="MACD 12-26-9 Bearish Crossover",
                    type_info="Short",
                    type=Optional[date],
                    function=lambda d: cross(d.MACD_12_26_9_SIGNAL, d.MACD_12_26_9),
                ),
                Signal(
                    name="MACD_12_26_9_ZERO_LINE_CROSS_UP",
                    description="MACD 12-26-9 Zero Line Cross Up",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value(d.MACD_12_26_9, 0),
                ),
                Signal(
                    name="MACD_12_26_9_ZERO_LINE_CROSS_DOWN",
                    description="MACD 12-26-9 Zero Line Cross Down",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value(d.MACD_12_26_9, 0, above=False),
                ),
            ],
        ),
        Indicator(
            name="RSI",
            description="Relative Strength Index",
            expected_columns=RSI.expected_columns,
            function=RSI.call,
            signals=[
                Signal(
                    name="RSI_BULLISH_CROSSOVER_30",
                    description="RSI Bullish Crossover",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value(d.RSI, 30),
                ),
                Signal(
                    name="RSI_BULLISH_CROSSOVER_40",
                    description="RSI Bullish Crossover 40",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value(d.RSI, 40),
                ),
                Signal(
                    name="RSI_BULLISH_CROSSOVER_45",
                    description="RSI Bullish Crossover 45",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value(d.RSI, 45),
                ),
                Signal(
                    name="RSI_BEARISH_CROSSOVER",
                    description="RSI Bearish Crossover",
                    type_info="Short",
                    type=Optional[date],
                    function=lambda d: cross_value(d.RSI, 70, above=False),
                ),
                Signal(
                    name="RSI_OVERSOLD",
                    description="RSI Oversold Signal",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: d[(d.RSI < 30) & (d.RSI > 0)].last_valid_index(),
                ),
                Signal(
                    name="RSI_OVERBOUGHT",
                    description="RSI Overbought Signal",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: d[
                        (d.RSI < 100) & (d.RSI > 70)
                    ].last_valid_index(),
                ),
                Signal(
                    name="RSI_NEUTRAL",
                    description="RSI Neutral Signal",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: d[
                        (d.RSI < 60) & (d.RSI > 40)
                    ].last_valid_index(),
                ),
                Signal(
                    name="RETURN_AFTER_RSI_CROSSOVER_45_PERIOD_90",
                    description="Percentile 30 return after RSI crossover 45 in the next 90 days",
                    type_info="Long",
                    type=Optional[float],
                    function=lambda d: compute_percentile_return_after_rsi_crossover(d),
                ),
            ],
        ),
        Indicator(
            name="STOCH",
            description="Stochastic",
            expected_columns=["SLOW_K", "SLOW_D"],
            function=STOCH.call,
            signals=[
                Signal(
                    name="STOCH_OVERSOLD",
                    description="Stoch Oversold Signal",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: d[
                        (d.SLOW_K < 20) & (d.SLOW_K > 0)
                    ].last_valid_index(),
                ),
                Signal(
                    name="STOCH_OVERBOUGHT",
                    description="Stoch Overbought Signal",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: d[
                        (d.SLOW_K < 100) & (d.SLOW_K > 80)
                    ].last_valid_index(),
                ),
            ],
        ),
        Indicator(
            name="MFI",
            description="Money Flow Index",
            expected_columns=["MFI"],
            function=MFI.call,
            signals=[
                Signal(
                    name="MFI_OVERSOLD",
                    description="MFI Oversold Signal",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: d[(d.MFI < 20)].last_valid_index(),
                ),
                Signal(
                    name="MFI_OVERBOUGHT",
                    description="MFI Overbought Signal",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: d[(d.MFI > 80)].last_valid_index(),
                ),
            ],
        ),
        Indicator(
            name="SMA",
            description="Money Flow Index",
            expected_columns=["SMA_50", "SMA_200"],
            function=SMA.call,
            signals=[
                Signal(
                    name="GOLDEN_CROSS",
                    description="Golden cross: SMA 50 crosses above SMA 200",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: cross(d.SMA_50, d.SMA_200),
                ),
                Signal(
                    name="DEATH_CROSS",
                    description="Death cross: SMA 50 crosses below SMA 200",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: cross(d.SMA_50, d.SMA_200, above=False),
                ),
                Signal(
                    name="MOMENTUM_TIME_SPAN",
                    description="Momentum time span",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: momentum(d),
                ),
            ],
        ),
        Indicator(
            name="PRICE",
            description="Price based indicators",
            expected_columns=PRICE.expected_columns,
            function=PRICE.call,
            signals=[
                Signal(
                    name="LOWER_THAN_200_DAY_HIGH",
                    description="Current price is lower than the 200-day high",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: d[
                        0.6 * d["200_DAY_HIGH"] > d.LAST_PRICE
                    ].last_valid_index(),
                ),
                Signal(
                    name="LOWER_THAN_20_DAY_HIGH",
                    description="Current price is lower than the 20-day high",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: d[
                        0.6 * d["20_DAY_HIGH"] > d.LAST_PRICE
                    ].last_valid_index(),
                ),
                Signal(
                    name="MEDIAN_WEEKLY_GROWTH",
                    description="Median weekly growth",
                    type_info="Oversold",
                    type=Optional[float],
                    function=lambda d: np.median(d.WEEKLY_GROWTH.unique()),
                ),
                Signal(
                    name="MEDIAN_MONTHLY_GROWTH",
                    description="Median monthly growth",
                    type_info="Oversold",
                    type=Optional[float],
                    function=lambda d: np.median(d.MONTHLY_GROWTH.unique()),
                ),
                Signal(
                    name="MEDIAN_YEARLY_GROWTH",
                    description="Median yearly growth",
                    type_info="Oversold",
                    type=Optional[float],
                    function=lambda d: np.median(d.YEARLY_GROWTH.unique()),
                ),
            ],
        ),
        Indicator(
            name="ROC",
            description="Rate Of Change",
            expected_columns=ROC.expected_columns,
            function=ROC.call,
            signals=[
                Signal(
                    name="MEDIAN_RATE_OF_CHANGE_1",
                    type_info="Value",
                    description="Median daily Rate of Change of the last 30 days",
                    type=Optional[float],
                    function=lambda d: np.median(d.ROC_1.tolist()[-30:]),
                ),
                Signal(
                    name="MEDIAN_RATE_OF_CHANGE_7_4",
                    type_info="Value",
                    description="Median weekly Rate of Change of the last 4 weeks",
                    type=Optional[float],
                    function=lambda d: np.median(d.ROC_7.tolist()[-4:]),
                ),
                Signal(
                    name="MEDIAN_RATE_OF_CHANGE_7_12",
                    type_info="Value",
                    description="Median weekly Rate of Change of the last 12 weeks",
                    type=Optional[float],
                    function=lambda d: np.median(d.ROC_7.tolist()[-12:]),
                ),
                Signal(
                    name="MEDIAN_RATE_OF_CHANGE_30",
                    type_info="Value",
                    description="Median monthly Rate of Change of the last 12 Months",
                    type=Optional[float],
                    function=lambda d: np.median(d.ROC_30.tolist()[-12:]),
                ),
                Signal(
                    name="RATE_OF_CHANGE_30",
                    type_info="Value",
                    description="30-day Rate of Change",
                    type=Optional[float],
                    function=lambda d: d.ROC_30.tolist()[-1],
                ),
                Signal(
                    name="RATE_OF_CHANGE_7",
                    type_info="Value",
                    description="7-day Rate of Change",
                    type=Optional[float],
                    function=lambda d: d.ROC_7.tolist()[-1],
                ),
                Signal(
                    name="MOMENTUM",
                    type_info="Value",
                    description="7-day Rate of Change",
                    type=Optional[float],
                    function=lambda d: d.MOM.iloc[-1],
                ),
            ],
        ),
        Indicator(
            name="ADOSC",
            description="Chaikin A/D Oscillator",
            expected_columns=["ADOSC", "ADOSC_SIGNAL"],
            function=ADOSC.call,
            signals=[
                Signal(
                    name="ADOSC_CROSSES_ABOVE_0",
                    type_info="Oversold",
                    description="Bullish momentum in money flow",
                    type=Optional[date],
                    function=lambda d: cross_value(d.ADOSC, 0, above=True),
                ),
                Signal(
                    name="POSITIVE_ADOSC_20_DAY_BREAKOUT",
                    type_info="Oversold",
                    description="20-day breakout confirmed by positive ADOSC",
                    type=Optional[date],
                    function=lambda d: d[
                        (d.ADOSC_SIGNAL == True)  # noqa: E712
                    ].last_valid_index(),
                ),
            ],
        ),
        Indicator(
            name="CANDLESTICKS",
            description="Candlestick Patterns",
            expected_columns=[
                "CDLMORNINGSTAR",
                "CDL3LINESTRIKE",
                "CDL3WHITESOLDIERS",
                "CDLABANDONEDBABY",
                "CDLTASUKIGAP",
                "CDLPIERCING",
                "CDLENGULFING",
            ],
            function=CANDLESTOCK_PATTERNS.call,
            signals=[
                Signal(
                    name="CDLMORNINGSTAR",
                    type_info="Long",
                    description="Morning Star Candlestick Pattern",
                    type=Optional[date],
                    function=lambda d: d[(d.CDLMORNINGSTAR == 100)].last_valid_index(),
                ),
                Signal(
                    name="CDL3LINESTRIKE",
                    description="3 Line Strike Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d[(d.CDL3LINESTRIKE == 100)].last_valid_index(),
                ),
                Signal(
                    name="CDL3WHITESOLDIERS",
                    description="3 White Soldiers Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d[
                        (d.CDL3WHITESOLDIERS == 100)
                    ].last_valid_index(),
                ),
                Signal(
                    name="CDLABANDONEDBABY",
                    description="Abandoned Baby Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d[
                        (d.CDLABANDONEDBABY == 100)
                    ].last_valid_index(),
                ),
                Signal(
                    name="CDLTASUKIGAP",
                    description="Tasukigap Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d[(d.CDLTASUKIGAP == 100)].last_valid_index(),
                ),
                Signal(
                    name="CDLPIERCING",
                    description="Piercing Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d[(d.CDLPIERCING == 100)].last_valid_index(),
                ),
                Signal(
                    name="CDLENGULFING",
                    description="Engulfing Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d[(d.CDLENGULFING == 100)].last_valid_index(),
                ),
            ],
        ),
    ]


class Indicators(BaseModel):
    indicators: List[Indicator] = Field(default_factory=indicators_factory)

    def compute(self, data: pd.DataFrame) -> None:
        for indicator in self.indicators:
            try:
                indicator.compute(data)
            except Exception as e:
                logger.error(f"Failed to compute indicator {indicator.name}: {e}")
                continue
            logger.info(
                f"Computed {indicator.name} with {len(indicator.signals)} signals"
            )

    def to_dict(self, data: pd.DataFrame) -> Dict[str, Any]:
        self.compute(data)
        res = {}
        for indicator in self.indicators:
            for signal in indicator.signals:
                res[signal.name.lower()] = (
                    signal.date if signal.is_date() else signal.value
                )
        return res

    def create_indicator_models(self) -> List[type[BaseModel]]:
        models = []
        for indicator in self.indicators:
            model_parameters = {}
            for signal in indicator.signals:
                range_ = {}
                if signal.range:
                    range_ = {"ge": signal.range[0], "le": signal.range[1]}
                model_parameters[signal.name.lower()] = (
                    signal.type,
                    Field(  # type: ignore
                        None,
                        **range_,
                        description=(
                            signal.description
                            or " ".join(signal.name.lower().capitalize().split("_"))
                        ),
                    ),
                )
            model = create_model(indicator.name, **model_parameters)  # type: ignore
            model._description = indicator.description
            models.append(model)
        return models


IndicatorModels = Indicators().create_indicator_models()



import datetime
from datetime import date
from typing import Literal, get_args, Any, Optional, List, Tuple, Type, Dict

from bearish.types import SeriesLength  # type: ignore
from pydantic import BaseModel, Field, ConfigDict
from pydantic import create_model
from pydantic.fields import FieldInfo

from bullish.analysis.analysis import (
    YearlyFundamentalAnalysis,
    QuarterlyFundamentalAnalysis,
    TechnicalAnalysisModels,
    TechnicalAnalysis,
    AnalysisView,
)

Industry = Literal[
    "Publishing",
    "Internet Retail",
    "Scientific & Technical Instruments",
    "Engineering & Construction",
    "Diagnostics & Research",
    "Software - Infrastructure",
    "Thermal Coal",
    "Software - Application",
    "Auto Manufacturers",
    "Farm Products",
    "Medical Devices",
    "Education & Training Services",
    "Auto Parts",
    "Specialty Chemicals",
    "Marine Shipping",
    "Biotechnology",
    "Real Estate Services",
    "Gold",
    "Entertainment",
    "Specialty Retail",
    "Utilities - Independent Power Producers",
    "Steel",
    "Mortgage Finance",
    "Communication Equipment",
    "Drug Manufacturers - Specialty & Generic",
    "Electronic Gaming & Multimedia",
    "Banks - Regional",
    "Oil & Gas E&P",
    "Travel Services",
    "Real Estate - Diversified",
    "Telecom Services",
    "Uranium",
    "Consulting Services",
    "Waste Management",
    "Agricultural Inputs",
    "Utilities - Diversified",
    "Auto & Truck Dealerships",
    "Confectioners",
    "Other Industrial Metals & Mining",
    "Beverages - Wineries & Distilleries",
    "Oil & Gas Midstream",
    "Recreational Vehicles",
    "Electrical Equipment & Parts",
    "Household & Personal Products",
    "Packaging & Containers",
    "REIT - Specialty",
    "Home Improvement Retail",
    "Electronic Components",
    "Asset Management",
    "Consumer Electronics",
    "Conglomerates",
    "Health Information Services",
    "Medical Instruments & Supplies",
    "Building Products & Equipment",
    "Information Technology Services",
    "Specialty Industrial Machinery",
    "Food Distribution",
    "Packaged Foods",
    "Rental & Leasing Services",
    "Medical Distribution",
    "Grocery Stores",
    "Advertising Agencies",
    "Beverages - Non - Alcoholic",
    "Apparel Manufacturing",
    "Oil & Gas Equipment & Services",
    "Coking Coal",
    "Industrial Distribution",
    "Restaurants",
    "Beverages - Brewers",
    "Chemicals",
    "Real Estate - Development",
    "Credit Services",
    "Tobacco",
    "Metal Fabrication",
    "Building Materials",
    "Residential Construction",
    "Specialty Business Services",
    "REIT - Hotel & Motel",
    "Internet Content & Information",
    "Lodging",
    "Furnishings, Fixtures & Appliances",
    "Airlines",
    "Computer Hardware",
    "Integrated Freight & Logistics",
    "Solar",
    "Capital Markets",
    "Leisure",
    "Airports & Air Services",
    "Aluminum",
    "Insurance Brokers",
    "Semiconductors",
    "REIT - Retail",
    "Luxury Goods",
    "Lumber & Wood Production",
    "REIT - Mortgage",
    "Semiconductor Equipment & Materials",
    "Aerospace & Defense",
    "Security & Protection Services",
    "Utilities - Renewable",
    "Utilities - Regulated Gas",
    "Apparel Retail",
    "Pollution & Treatment Controls",
    "Broadcasting",
    "Resorts & Casinos",
    "Other Precious Metals & Mining",
    "Financial Data & Stock Exchanges",
    "Footwear & Accessories",
    "Medical Care Facilities",
    "Electronics & Computer Distribution",
    "Gambling",
    "Tools & Accessories",
    "Insurance - Property & Casualty",
    "Utilities - Regulated Water",
    "Insurance - Specialty",
    "Personal Services",
    "Pharmaceutical Retailers",
    "Farm & Heavy Construction Machinery",
    "Utilities - Regulated Electric",
    "Department Stores",
    "Staffing & Employment Services",
    "Textile Manufacturing",
    "Silver",
    "REIT - Industrial",
    "REIT - Diversified",
    "Copper",
    "Business Equipment & Supplies",
    "Infrastructure Operations",
    "Trucking",
    "Insurance - Reinsurance",
    "Insurance - Diversified",
    "Drug Manufacturers - General",
    "Oil & Gas Drilling",
    "Banks - Diversified",
    "REIT - Residential",
    "Oil & Gas Refining & Marketing",
    "Shell Companies",
    "Financial Conglomerates",
    "Paper & Paper Products",
    "Insurance - Life",
    "REIT - Office",
    "Railroads",
    "Oil & Gas Integrated",
    "Healthcare Plans",
    "REIT - Healthcare Facilities",
    "Discount Stores",
]

IndustryGroup = Literal[
    "publishing",
    "internet-retail",
    "scientific-technical-instruments",
    "engineering-construction",
    "diagnostics-research",
    "software-infrastructure",
    "thermal-coal",
    "software-application",
    "auto-manufacturers",
    "farm-products",
    "medical-devices",
    "education-training-services",
    "auto-parts",
    "specialty-chemicals",
    "marine-shipping",
    "biotechnology",
    "real-estate-services",
    "gold",
    "entertainment",
    "specialty-retail",
    "utilities-independent-power-producers",
    "steel",
    "mortgage-finance",
    "communication-equipment",
    "drug-manufacturers-specialty-generic",
    "electronic-gaming-multimedia",
    "banks-regional",
    "oil-gas-e-p",
    "travel-services",
    "real-estate-diversified",
    "telecom-services",
    "uranium",
    "consulting-services",
    "waste-management",
    "agricultural-inputs",
    "utilities-diversified",
    "auto-truck-dealerships",
    "confectioners",
    "other-industrial-metals-mining",
    "beverages-wineries-distilleries",
    "oil-gas-midstream",
    "recreational-vehicles",
    "electrical-equipment-parts",
    "household-personal-products",
    "packaging-containers",
    "reit-specialty",
    "home-improvement-retail",
    "electronic-components",
    "asset-management",
    "consumer-electronics",
    "conglomerates",
    "health-information-services",
    "medical-instruments-supplies",
    "building-products-equipment",
    "information-technology-services",
    "specialty-industrial-machinery",
    "food-distribution",
    "packaged-foods",
    "rental-leasing-services",
    "medical-distribution",
    "grocery-stores",
    "advertising-agencies",
    "beverages-non-alcoholic",
    "apparel-manufacturing",
    "oil-gas-equipment-services",
    "coking-coal",
    "industrial-distribution",
    "restaurants",
    "beverages-brewers",
    "chemicals",
    "real-estate-development",
    "credit-services",
    "tobacco",
    "metal-fabrication",
    "building-materials",
    "residential-construction",
    "specialty-business-services",
    "reit-hotel-motel",
    "internet-content-information",
    "lodging",
    "furnishings-fixtures-appliances",
    "airlines",
    "computer-hardware",
    "integrated-freight-logistics",
    "solar",
    "capital-markets",
    "leisure",
    "airports-air-services",
    "aluminum",
    "insurance-brokers",
    "semiconductors",
    "reit-retail",
    "luxury-goods",
    "lumber-wood-production",
    "reit-mortgage",
    "semiconductor-equipment-materials",
    "aerospace-defense",
    "security-protection-services",
    "utilities-renewable",
    "utilities-regulated-gas",
    "apparel-retail",
    "pollution-treatment-controls",
    "broadcasting",
    "resorts-casinos",
    "other-precious-metals-mining",
    "financial-data-stock-exchanges",
    "footwear-accessories",
    "medical-care-facilities",
    "electronics-computer-distribution",
    "gambling",
    "tools-accessories",
    "insurance-property-casualty",
    "utilities-regulated-water",
    "insurance-specialty",
    "personal-services",
    "pharmaceutical-retailers",
    "farm-heavy-construction-machinery",
    "utilities-regulated-electric",
    "department-stores",
    "staffing-employment-services",
    "textile-manufacturing",
    "silver",
    "reit-industrial",
    "reit-diversified",
    "copper",
    "business-equipment-supplies",
    "infrastructure-operations",
    "trucking",
    "insurance-reinsurance",
    "insurance-diversified",
    "drug-manufacturers-general",
    "oil-gas-drilling",
    "banks-diversified",
    "reit-residential",
    "oil-gas-refining-marketing",
    "shell-companies",
    "financial-conglomerates",
    "paper-paper-products",
    "insurance-life",
    "reit-office",
    "railroads",
    "oil-gas-integrated",
    "healthcare-plans",
    "reit-healthcare-facilities",
    "discount-stores",
]

Sector = Literal[
    "Communication Services",
    "Consumer Cyclical",
    "Technology",
    "Industrials",
    "Healthcare",
    "Energy",
    "Consumer Defensive",
    "Basic Materials",
    "Real Estate",
    "Utilities",
    "Financial Services",
    "Conglomerates",
]

Country = Literal[
    "Australia",
    "China",
    "Japan",
    "United kingdom",
    "United states",
    "Poland",
    "Switzerland",
    "Canada",
    "Greece",
    "Spain",
    "Germany",
    "Indonesia",
    "Belgium",
    "France",
    "Netherlands",
    "British virgin islands",
    "Italy",
    "Hungary",
    "Austria",
    "Finland",
    "Sweden",
    "Bermuda",
    "Taiwan",
    "Israel",
    "Ukraine",
    "Singapore",
    "Jersey",
    "Ireland",
    "Luxembourg",
    "Cyprus",
    "Cayman islands",
    "Norway",
    "Denmark",
    "Hong kong",
    "New zealand",
    "Kazakhstan",
    "Nigeria",
    "Argentina",
    "Brazil",
    "Czech republic",
    "Mauritius",
    "South africa",
    "India",
    "Mexico",
    "Mongolia",
    "Slovenia",
    "Thailand",
    "Malaysia",
    "Costa rica",
    "Isle of man",
    "Egypt",
    "Turkey",
    "United arab emirates",
    "Colombia",
    "Gibraltar",
    "Malta",
    "Liechtenstein",
    "Guernsey",
    "Peru",
    "Estonia",
    "French guiana",
    "Portugal",
    "Uruguay",
    "Chile",
    "Martinique",
    "Monaco",
    "Panama",
    "Papua new guinea",
    "South korea",
    "Macau",
    "Gabon",
    "Romania",
    "Senegal",
    "Morocco",
    "Jordan",
    "Lithuania",
    "Dominican republic",
    "Reunion",
    "Zambia",
    "Cambodia",
    "Myanmar",
    "Bahamas",
    "Philippines",
    "Bangladesh",
    "Latvia",
    "Vietnam",
    "Iceland",
    "Azerbaijan",
    "Georgia",
    "Liberia",
    "Kenya",
]
SIZE_RANGE = 2


def _get_type(name: str, info: FieldInfo) -> Tuple[Any, Any]:
    alias = info.alias or " ".join(name.capitalize().split("_")).strip()
    if info.annotation == Optional[float]:  # type: ignore
        ge = next((item.ge for item in info.metadata if hasattr(item, "ge")), 0)
        le = next((item.le for item in info.metadata if hasattr(item, "le")), 100)
        default = [ge, le]
        return (
            Optional[List[float]],
            Field(default=default, alias=alias, description=info.description),
        )
    elif info.annotation == Optional[date]:  # type: ignore
        le = date.today()
        ge = le - datetime.timedelta(days=30 * 2)  # 30 days * 12 months
        return (
            List[date],
            Field(default=[ge, le], alias=alias, description=info.description),
        )
    else:
        raise NotImplementedError


FUNDAMENTAL_ANALYSIS_GROUP = ["income", "cash_flow", "eps"]


def _get_fundamental_analysis_boolean_fields() -> List[str]:
    return [
        name
        for name, info in {
            **YearlyFundamentalAnalysis.model_fields,
            **QuarterlyFundamentalAnalysis.model_fields,
        }.items()
        if info.annotation == Optional[bool]
    ]


def _get_technical_analysis_float_fields() -> List[str]:
    return [
        name
        for name, info in (
            TechnicalAnalysis.model_fields | AnalysisView.model_fields
        ).items()
        if info.annotation == Optional[float]
    ]


def get_boolean_field_group(group: str) -> List[str]:
    groups = FUNDAMENTAL_ANALYSIS_GROUP.copy()
    groups.remove(group)
    return [
        name
        for name in _get_fundamental_analysis_boolean_fields()
        if group in name and not any(g in name for g in groups)
    ]


INCOME_GROUP = get_boolean_field_group("income")
CASH_FLOW_GROUP = get_boolean_field_group("cash_flow")
EPS_GROUP = get_boolean_field_group("eps")
PROPERTIES_GROUP = list(
    set(_get_fundamental_analysis_boolean_fields()).difference(
        {*INCOME_GROUP, *CASH_FLOW_GROUP, *EPS_GROUP}
    )
)

GROUP_MAPPING: Dict[str, List[str]] = {
    "income": INCOME_GROUP,
    "cash_flow": CASH_FLOW_GROUP,
    "eps": EPS_GROUP,
    "properties": PROPERTIES_GROUP,
    "country": list(get_args(Country)),
    "industry": list(get_args(Industry)),
    "industry_group": list(get_args(IndustryGroup)),
    "sector": list(get_args(Sector)),
    "symbol": [],
    "order_by_asc": _get_technical_analysis_float_fields(),
    "order_by_desc": _get_technical_analysis_float_fields(),
}


def _create_fundamental_analysis_models() -> List[Type[BaseModel]]:
    models = []
    boolean_fields = {
        "income": (Optional[List[str]], Field(default=None, description="Income")),
        "cash_flow": (
            Optional[List[str]],
            Field(default=None, description="Cash flow"),
        ),
        "eps": (
            Optional[List[str]],
            Field(default=None, description="Earnings per share"),
        ),
        "properties": (
            Optional[List[str]],
            Field(default=None, description="General properties"),
        ),
    }
    yearly_fields = {
        name: _get_type(name, info)
        for name, info in YearlyFundamentalAnalysis.model_fields.items()
        if info.annotation != Optional[bool]  # type: ignore
    }
    quarterly_fields = {
        name: _get_type(name, info)
        for name, info in QuarterlyFundamentalAnalysis.model_fields.items()
        if info.annotation != Optional[bool]
    }
    for property in [
        (boolean_fields, "Selection filter", "SelectionFilter"),
        (yearly_fields, "Yearly properties", "YearlyFilter"),
        (quarterly_fields, "Quarterly properties", "QuarterlyFilter"),
    ]:
        model_ = create_model(  # type: ignore
            property[-1],
            __config__=ConfigDict(populate_by_name=True),
            **property[0],
        )
        model_._description = property[1]
        models.append(model_)

    return models


def create_technical_analysis_models() -> List[Type[BaseModel]]:
    models = []
    for model in TechnicalAnalysisModels:
        model_ = create_model(  # type: ignore
            f"{model.__name__}Filter",  # type: ignore
            __config__=ConfigDict(populate_by_name=True),
            **{
                name: _get_type(name, info) for name, info in model.model_fields.items()  # type: ignore
            },
        )

        model_._description = model._description  # type: ignore
        models.append(model_)
    return models


TechnicalAnalysisFilters = create_technical_analysis_models()
FundamentalAnalysisFilters = _create_fundamental_analysis_models()


class GeneralFilter(BaseModel):
    country: Optional[List[str]] = None
    order_by_asc: Optional[str] = None
    order_by_desc: Optional[str] = None
    industry: Optional[List[str]] = None
    industry_group: Optional[List[str]] = None
    sector: Optional[List[str]] = None
    symbol: Optional[List[str]] = None
    market_capitalization: Optional[List[float]] = Field(default=[5e8, 1e12])
    price_per_earning_ratio: Optional[List[float]] = Field(default=[0.0, 1000.0])


class FilterQuery(GeneralFilter, *TechnicalAnalysisFilters, *FundamentalAnalysisFilters):  # type: ignore

    def valid(self) -> bool:
        return any(
            bool(v)
            for _, v in self.model_dump(
                exclude_defaults=True, exclude_unset=True
            ).items()
        )

    def to_query(self) -> str:
        parameters = self.model_dump(exclude_defaults=True, exclude_unset=True)
        query = []
        order_by_desc = ""
        order_by_asc = ""
        for parameter, value in parameters.items():
            if not value:
                continue

            if (
                isinstance(value, list)
                and all(isinstance(item, str) for item in value)
                and parameter not in GeneralFilter.model_fields
            ):
                query.append(" AND ".join([f"{v}=1" for v in value]))
            elif (
                isinstance(value, str) and bool(value) and parameter == "order_by_desc"
            ):
                order_by_desc = f"ORDER BY {value} DESC"
            elif isinstance(value, str) and bool(value) and parameter == "order_by_asc":
                order_by_asc = f"ORDER BY {value} ASC"
            elif (
                isinstance(value, list)
                and len(value) == SIZE_RANGE
                and all(isinstance(item, (int, float)) for item in value)
            ):
                query.append(f"{parameter} BETWEEN {value[0]} AND {value[1]}")
            elif (
                isinstance(value, list)
                and len(value) == SIZE_RANGE
                and all(isinstance(item, date) for item in value)
            ):
                query.append(f"{parameter} BETWEEN '{value[0]}' AND '{value[1]}'")
            elif (
                isinstance(value, list)
                and all(isinstance(item, str) for item in value)
                and parameter in GeneralFilter.model_fields
            ):
                general_filters = [f"'{v}'" for v in value]
                query.append(f"{parameter} IN ({', '.join(general_filters)})")
            else:
                raise NotImplementedError
        query_ = " AND ".join(query)
        return f"{query_} {order_by_desc.strip()} {order_by_asc.strip()}".strip()


class FilterQueryStored(FilterQuery): ...


class FilterUpdate(BaseModel):
    window_size: SeriesLength = Field("5d")
    data_age_in_days: int = 1
    update_financials: bool = False
    update_analysis_only: bool = False


class FilteredResults(BaseModel):
    name: str
    filter_query: FilterQueryStored
    symbols: list[str] = Field(
        default_factory=list, description="List of filtered tickers."
    )
